library flutter_puzzle;

import 'dart:io';
import 'dart:math';
import 'package:flutter/material.dart';

//region puzzle piece widget

class _PuzzlePiece extends StatefulWidget {
  final Widget image;
  final Size imageSize;
  final double top;
  final double left;
  final Function bringToTop;
  final Function sendToBack;
  final Function isDone;

  final _PieceData data;

  _PuzzlePiece({
    Key key,
    @required this.image,
    @required this.imageSize,
    @required this.top,
    @required this.left,
    @required this.data,
    @required this.bringToTop,
    @required this.sendToBack,
    @required this.isDone,
  }) : super(key: key);

  @override
  _PuzzlePieceState createState() {
    return new _PuzzlePieceState();
  }
}

class _PuzzlePieceState extends State<_PuzzlePiece> {
  double _top;
  double _left;
  bool _isMovable = true;

  @override
  Widget build(BuildContext context) {
    final imageWidth = widget.imageSize.width;
    final imageHeight = widget.imageSize.height;
    final pieceWidth = imageWidth / widget.data.maxCol;
    final pieceHeight = imageHeight / widget.data.maxRow;

    if (_top == null) {
      _top = Random().nextInt((imageHeight - pieceHeight).ceil()).toDouble();
      _top -= widget.data.row * pieceHeight;
    }
    if (_left == null) {
      _left = Random().nextInt((imageWidth - pieceWidth).ceil()).toDouble();
      _left -= widget.data.col * pieceWidth;
    }

    return Positioned(
      top: _top,
      left: _left,
      width: imageWidth,
      child: GestureDetector(
        onTap: () {
          if (_isMovable) {
            widget.bringToTop(widget);
          }
        },
        onPanStart: (_) {
          if (_isMovable) {
            widget.bringToTop(widget);
          }
        },
        onPanUpdate: (dragUpdateDetails) {
          if (_isMovable) {
            setState(() {
              _top += dragUpdateDetails.delta.dy;
              _left += dragUpdateDetails.delta.dx;

              if (_inPosition(widget.top, widget.left)) {
                _top = widget.top;
                _left = widget.left;
                _isMovable = false;
                widget.sendToBack(widget);
                widget.isDone();
              }
            });
          }
        },
        child: ClipPath(
          child: CustomPaint(
            foregroundPainter: _PuzzlePiecePainter(data: widget.data),
            child: widget.image,
            ),
          clipper: _PuzzlePieceClipper(data: widget.data),
          ),
        ),
      );
  }

  bool _inPosition(double _snapY, double _snapX) {
    return (_snapY - 10 < _top) &&
        (_top < _snapY + 10) &&
        (_snapX - 10 < _left) &&
        (_left < _snapX + 10);
  }
}

//endregion

//region puzzle piece builder

class _PieceData {
  final int row;
  final int col;
  final int maxRow;
  final int maxCol;
  final double stroke;

  _PieceData({
    @required this.row,
    @required this.col,
    @required this.maxRow,
    @required this.maxCol,
    @required this.stroke,
  });
}

class _PuzzlePiecePainter extends CustomPainter {
  final _PieceData data;

  _PuzzlePiecePainter({@required this.data});

  @override
  void paint(Canvas canvas, Size size) {
    Paint style = Paint();
    style.color = Colors.white;
    style.style = PaintingStyle.stroke;
    style.strokeWidth = data.stroke;

    final Path path = _PuzzlePiecePath(size: size, data: this.data).getPath();
    return canvas.drawPath(path, style);
  }

  @override
  bool shouldRepaint(CustomPainter oldDelegate) => false;
}

class _PuzzlePieceClipper extends CustomClipper<Path> {
  final _PieceData data;

  _PuzzlePieceClipper({@required this.data});

  @override
  Path getClip(Size size) {
    return _PuzzlePiecePath(size: size, data: this.data).getPath();
  }

  @override
  bool shouldReclip(CustomClipper<Path> oldClipper) => false;
}

class _PuzzlePiecePath {
  Path _path;

  _PuzzlePiecePath({@required _PieceData data, @required Size size}) {
    this._path = _getPiecePath(size, data);
  }

  Path getPath() => this._path;

  Path _getPiecePath(Size size, _PieceData data) {
    final width = size.width / data.maxCol;
    final height = size.height / data.maxRow;
    final offsetX = data.col * width;
    final offsetY = data.row * height;
    final verticalBumpSize = width / 4;
    final horizontalBumpSize = height / 4;

    var path = Path();
    path.moveTo(offsetX, offsetY);

    if (data.row == 0) {
      // top side piece
      path.lineTo(offsetX + width, offsetY);
    } else {
      // top bump
      path.lineTo(offsetX + width / 3, offsetY);
      path.cubicTo(
        offsetX + width / 6,
        offsetY - verticalBumpSize,
        offsetX + width / 6 * 5,
        offsetY - verticalBumpSize,
        offsetX + width / 3 * 2,
        offsetY,
        );
      path.lineTo(offsetX + width, offsetY);
    }

    if (data.col == data.maxCol - 1) {
      // right side piece
      path.lineTo(offsetX + width, offsetY + height);
    } else {
      // right bump
      path.lineTo(offsetX + width, offsetY + height / 3);
      path.cubicTo(
        offsetX + width - horizontalBumpSize,
        offsetY + height / 6,
        offsetX + width - horizontalBumpSize,
        offsetY + height / 6 * 5,
        offsetX + width,
        offsetY + height / 3 * 2,
        );
      path.lineTo(offsetX + width, offsetY + height);
    }

    if (data.row == data.maxRow - 1) {
      // bottom side piece
      path.lineTo(offsetX, offsetY + height);
    } else {
      // bottom bump
      path.lineTo(offsetX + width / 3 * 2, offsetY + height);
      path.cubicTo(
        offsetX + width / 6 * 5,
        offsetY + height - verticalBumpSize,
        offsetX + width / 6,
        offsetY + height - verticalBumpSize,
        offsetX + width / 3,
        offsetY + height,
        );
      path.lineTo(offsetX, offsetY + height);
    }

    if (data.col == 0) {
      // left side piece
      path.close();
    } else {
      // left bump
      path.lineTo(offsetX, offsetY + height / 3 * 2);
      path.cubicTo(
        offsetX - horizontalBumpSize,
        offsetY + height / 6 * 5,
        offsetX - horizontalBumpSize,
        offsetY + height / 6,
        offsetX,
        offsetY + height / 3,
        );
      path.close();
    }

    return path;
  }
}

//endregion

//region puzzle widget

class PuzzleWidget extends StatefulWidget {
  final int rowsCount;
  final int colsCount;
  final double opacity;
  final double strokeWidth;
  final double top;
  final double left;
  final File file;
  final Size imageSize;
  final Function callback;

  PuzzleWidget({
    Key key,
    @required this.rowsCount,
    @required this.colsCount,
    @required this.opacity,
    @required this.strokeWidth,
    @required this.top,
    @required this.left,
    @required this.file,
    @required this.imageSize,
    @required this.callback,
  }) : super(key: key);

  @override
  _PuzzleWidgetState createState() => _PuzzleWidgetState();
}

class _PuzzleWidgetState extends State<PuzzleWidget> {
  List<Widget> _pieces = [];
  Widget _background;
  int _doneCount = 0;

  //region puzzle piece callbacks

  void bringToTop(Widget widget) {
    setState(() {
      _pieces.remove(widget);
      _pieces.add(widget);
    });
  }

  void sendToBack(Widget widget) {
    setState(() {
      _pieces.remove(widget);
      _pieces.insert(1, widget);
    });
  }

  void isDone() {
    int allCount = widget.rowsCount * widget.colsCount;

    this._doneCount++;

    if (_doneCount == allCount) {
      setState(() {
        _pieces.clear();
        _background = drawBackground(1);
        _pieces.insert(0, _background);
        Future.delayed(Duration(seconds: 1), () => widget.callback());
      });
    }
  }

  //endregion

  //region widget build

  @override
  Widget build(BuildContext context) {
    return Scaffold(
      body: Container(
        child: Center(
            child: widget.file == null ? Text('No image selected.') : Stack(children: _pieces)),
        ),
      );
  }

  Widget drawBackground(double opacity) => Positioned(
    top: widget.top,
    left: widget.left,
    width: widget.imageSize.width,
    height: widget.imageSize.height,
    child: Opacity(
      opacity: opacity,
      child: Image.file(
        widget.file,
        width: widget.imageSize.width,
        height: widget.imageSize.height,
        fit: BoxFit.cover,
        ),
      ),
    );

  //endregion

  @override
  void initState() {
    super.initState();
    if (widget.file != null) {
      Widget image = Image.file(
        widget.file,
        width: widget.imageSize.width,
        height: widget.imageSize.height,
        fit: BoxFit.cover,
        );

      for (int rowIndex = 0; rowIndex < widget.rowsCount; rowIndex++) {
        for (int columnIndex = 0; columnIndex < widget.colsCount; columnIndex++) {
          setState(() {
            _PieceData pieceData = _PieceData(
              row: rowIndex,
              col: columnIndex,
              maxRow: widget.rowsCount,
              maxCol: widget.colsCount,
              stroke: widget.strokeWidth,
              );

            _pieces.add(_PuzzlePiece(
              key: GlobalKey(),
              image: image,
              data: pieceData,
              bringToTop: this.bringToTop,
              sendToBack: this.sendToBack,
              imageSize: widget.imageSize,
              top: widget.top,
              left: widget.left,
              isDone: this.isDone,
              ));
          });
        }
      }

      _background = drawBackground(0.5);

      setState(() => _pieces.insert(0, _background));
    }
  }
}

//endregion
